package jdk.internal.org.objectweb.asm.tree.analysis;

import jdk.internal.org.objectweb.asm.tree.MultiANewArrayInsnNode;
import jdk.internal.org.objectweb.asm.tree.InvokeDynamicInsnNode;
import jdk.internal.org.objectweb.asm.Type;
import jdk.internal.org.objectweb.asm.tree.MethodInsnNode;
import java.util.ArrayList;
import jdk.internal.org.objectweb.asm.tree.IincInsnNode;
import jdk.internal.org.objectweb.asm.tree.VarInsnNode;
import jdk.internal.org.objectweb.asm.tree.AbstractInsnNode;

public class Frame<V extends Value>
{
    private V returnValue;
    private V[] values;
    private int locals;
    private int top;
    
    public Frame(int locals, int n) {
        this.values = (Value[])new Value[locals + n];
        this.locals = locals;
    }
    
    public Frame(Frame<? extends V> frame) {
        this(frame.locals, frame.values.length - frame.locals);
        this.init(frame);
    }
    
    public Frame<V> init(Frame<? extends V> frame) {
        this.returnValue = frame.returnValue;
        System.arraycopy(frame.values, 0, this.values, 0, this.values.length);
        this.top = frame.top;
        return this;
    }
    
    public void setReturn(V returnValue) {
        this.returnValue = returnValue;
    }
    
    public int getLocals() {
        return this.locals;
    }
    
    public int getMaxStackSize() {
        return this.values.length - this.locals;
    }
    
    public V getLocal(int n) throws IndexOutOfBoundsException {
        if (n >= this.locals) {
            throw new IndexOutOfBoundsException("Trying to access an inexistant local variable");
        }
        return this.values[n];
    }
    
    public void setLocal(int i, V v) throws IndexOutOfBoundsException {
        if (i >= this.locals) {
            throw new IndexOutOfBoundsException("Trying to access an inexistant local variable " + i);
        }
        this.values[i] = v;
        return;
    }
    
    public int getStackSize() {
        return this.top;
    }
    
    public V getStack(int n) throws IndexOutOfBoundsException {
        return this.values[n + this.locals];
    }
    
    public void clearStack() {
        this.top = 0;
    }
    
    public V pop() throws IndexOutOfBoundsException {
        if (this.top == 0) {
            throw new IndexOutOfBoundsException("Cannot pop operand off an empty stack.");
        }
        V[] values = this.values;
        int top = this.top - 1;
        this.top = top;
        return values[top + this.locals];
    }
    
    public void push(V v) throws IndexOutOfBoundsException {
        if (this.top + this.locals >= this.values.length) {
            throw new IndexOutOfBoundsException("Insufficient maximum stack size.");
        }
        this.values[this.top++ + this.locals] = v;
        return;
    }
    
    public void execute(AbstractInsnNode abstractInsnNode, Interpreter<V> interpreter) throws AnalyzerException {
        switch (abstractInsnNode.getOpcode()) {
            case 0: {
                break;
            }
            case 1:
            case 2:
            case 3:
            case 4:
            case 5:
            case 6:
            case 7:
            case 8:
            case 9:
            case 10:
            case 11:
            case 12:
            case 13:
            case 14:
            case 15:
            case 16:
            case 17:
            case 18: {
                this.push(interpreter.newOperation(abstractInsnNode));
                break;
            }
            case 21:
            case 22:
            case 23:
            case 24:
            case 25: {
                this.push(interpreter.copyOperation(abstractInsnNode, this.getLocal(((VarInsnNode)abstractInsnNode).var)));
                break;
            }
            case 46:
            case 47:
            case 48:
            case 49:
            case 50:
            case 51:
            case 52:
            case 53: {
                this.push(interpreter.binaryOperation(abstractInsnNode, this.pop(), this.pop()));
                break;
            }
            case 54:
            case 55:
            case 56:
            case 57:
            case 58: {
                Value copyOperation = interpreter.copyOperation(abstractInsnNode, this.pop());
                int var = ((VarInsnNode)abstractInsnNode).var;
                this.setLocal(var, copyOperation);
                if (copyOperation.getSize() == 2) {
                    this.setLocal(var + 1, interpreter.newValue(null));
                }
                if (var > 0) {
                    Value local = this.getLocal(var - 1);
                    if (local != null && local.getSize() == 2) {
                        this.setLocal(var - 1, interpreter.newValue(null));
                    }
                    break;
                }
                else {
                    break;
                }
            }
            case 79:
            case 80:
            case 81:
            case 82:
            case 83:
            case 84:
            case 85:
            case 86: {
                interpreter.ternaryOperation(abstractInsnNode, this.pop(), this.pop(), this.pop());
                break;
            }
            case 87: {
                if (this.pop().getSize() == 2) {
                    throw new AnalyzerException(abstractInsnNode, "Illegal use of POP");
                }
                break;
            }
            case 88: {
                if (this.pop().getSize() == 1 && this.pop().getSize() != 1) {
                    throw new AnalyzerException(abstractInsnNode, "Illegal use of POP2");
                }
                break;
            }
            case 89: {
                Value pop = this.pop();
                if (pop.getSize() != 1) {
                    throw new AnalyzerException(abstractInsnNode, "Illegal use of DUP");
                }
                this.push(pop);
                this.push(interpreter.copyOperation(abstractInsnNode, pop));
                break;
            }
            case 90: {
                Value pop2 = this.pop();
                Value pop3 = this.pop();
                if (pop2.getSize() != 1 || pop3.getSize() != 1) {
                    throw new AnalyzerException(abstractInsnNode, "Illegal use of DUP_X1");
                }
                this.push(interpreter.copyOperation(abstractInsnNode, pop2));
                this.push(pop3);
                this.push(pop2);
                break;
            }
            case 91: {
                Value pop4 = this.pop();
                if (pop4.getSize() == 1) {
                    Value pop5 = this.pop();
                    if (pop5.getSize() == 1) {
                        Value pop6 = this.pop();
                        if (pop6.getSize() == 1) {
                            this.push(interpreter.copyOperation(abstractInsnNode, pop4));
                            this.push(pop6);
                            this.push(pop5);
                            this.push(pop4);
                            break;
                        }
                    }
                    else {
                        this.push(interpreter.copyOperation(abstractInsnNode, pop4));
                        this.push(pop5);
                        this.push(pop4);
                        break;
                    }
                }
                throw new AnalyzerException(abstractInsnNode, "Illegal use of DUP_X2");
            }
            case 92: {
                Value pop7 = this.pop();
                if (pop7.getSize() == 1) {
                    Value pop8 = this.pop();
                    if (pop8.getSize() == 1) {
                        this.push(pop8);
                        this.push(pop7);
                        this.push(interpreter.copyOperation(abstractInsnNode, pop8));
                        this.push(interpreter.copyOperation(abstractInsnNode, pop7));
                        break;
                    }
                    throw new AnalyzerException(abstractInsnNode, "Illegal use of DUP2");
                }
                else {
                    this.push(pop7);
                    this.push(interpreter.copyOperation(abstractInsnNode, pop7));
                    break;
                }
            }
            case 93: {
                Value pop9 = this.pop();
                if (pop9.getSize() == 1) {
                    Value pop11 = this.pop();
                    if (pop11.getSize() == 1) {
                        Value pop12 = this.pop();
                        if (pop12.getSize() == 1) {
                            this.push(interpreter.copyOperation(abstractInsnNode, pop11));
                            this.push(interpreter.copyOperation(abstractInsnNode, pop9));
                            this.push(pop12);
                            this.push(pop11);
                            this.push(pop9);
                            break;
                        }
                    }
                }
                else {
                    Value pop10 = this.pop();
                    if (pop10.getSize() == 1) {
                        this.push(interpreter.copyOperation(abstractInsnNode, pop9));
                        this.push(pop10);
                        this.push(pop9);
                        break;
                    }
                }
                throw new AnalyzerException(abstractInsnNode, "Illegal use of DUP2_X1");
            }
            case 94: {
                Value pop13 = this.pop();
                if (pop13.getSize() == 1) {
                    Value pop16 = this.pop();
                    if (pop16.getSize() == 1) {
                        Value pop17 = this.pop();
                        if (pop17.getSize() == 1) {
                            Value pop18 = this.pop();
                            if (pop18.getSize() == 1) {
                                this.push(interpreter.copyOperation(abstractInsnNode, pop16));
                                this.push(interpreter.copyOperation(abstractInsnNode, pop13));
                                this.push(pop18);
                                this.push(pop17);
                                this.push(pop16);
                                this.push(pop13);
                                break;
                            }
                        }
                        else {
                            this.push(interpreter.copyOperation(abstractInsnNode, pop16));
                            this.push(interpreter.copyOperation(abstractInsnNode, pop13));
                            this.push(pop17);
                            this.push(pop16);
                            this.push(pop13);
                            break;
                        }
                    }
                }
                else {
                    Value pop14 = this.pop();
                    if (pop14.getSize() == 1) {
                        Value pop15 = this.pop();
                        if (pop15.getSize() == 1) {
                            this.push(interpreter.copyOperation(abstractInsnNode, pop13));
                            this.push(pop15);
                            this.push(pop14);
                            this.push(pop13);
                            break;
                        }
                    }
                    else {
                        this.push(interpreter.copyOperation(abstractInsnNode, pop13));
                        this.push(pop14);
                        this.push(pop13);
                        break;
                    }
                }
                throw new AnalyzerException(abstractInsnNode, "Illegal use of DUP2_X2");
            }
            case 95: {
                Value pop19 = this.pop();
                Value pop20 = this.pop();
                if (pop20.getSize() != 1 || pop19.getSize() != 1) {
                    throw new AnalyzerException(abstractInsnNode, "Illegal use of SWAP");
                }
                this.push(interpreter.copyOperation(abstractInsnNode, pop19));
                this.push(interpreter.copyOperation(abstractInsnNode, pop20));
                break;
            }
            case 96:
            case 97:
            case 98:
            case 99:
            case 100:
            case 101:
            case 102:
            case 103:
            case 104:
            case 105:
            case 106:
            case 107:
            case 108:
            case 109:
            case 110:
            case 111:
            case 112:
            case 113:
            case 114:
            case 115: {
                this.push(interpreter.binaryOperation(abstractInsnNode, this.pop(), this.pop()));
                break;
            }
            case 116:
            case 117:
            case 118:
            case 119: {
                this.push(interpreter.unaryOperation(abstractInsnNode, this.pop()));
                break;
            }
            case 120:
            case 121:
            case 122:
            case 123:
            case 124:
            case 125:
            case 126:
            case 127:
            case 128:
            case 129:
            case 130:
            case 131: {
                this.push(interpreter.binaryOperation(abstractInsnNode, this.pop(), this.pop()));
                break;
            }
            case 132: {
                int var2 = ((IincInsnNode)abstractInsnNode).var;
                this.setLocal(var2, interpreter.unaryOperation(abstractInsnNode, this.getLocal(var2)));
                break;
            }
            case 133:
            case 134:
            case 135:
            case 136:
            case 137:
            case 138:
            case 139:
            case 140:
            case 141:
            case 142:
            case 143:
            case 144:
            case 145:
            case 146:
            case 147: {
                this.push(interpreter.unaryOperation(abstractInsnNode, this.pop()));
                break;
            }
            case 148:
            case 149:
            case 150:
            case 151:
            case 152: {
                this.push(interpreter.binaryOperation(abstractInsnNode, this.pop(), this.pop()));
                break;
            }
            case 153:
            case 154:
            case 155:
            case 156:
            case 157:
            case 158: {
                interpreter.unaryOperation(abstractInsnNode, this.pop());
                break;
            }
            case 159:
            case 160:
            case 161:
            case 162:
            case 163:
            case 164:
            case 165:
            case 166: {
                interpreter.binaryOperation(abstractInsnNode, this.pop(), this.pop());
                break;
            }
            case 167: {
                break;
            }
            case 168: {
                this.push(interpreter.newOperation(abstractInsnNode));
                break;
            }
            case 169: {
                break;
            }
            case 170:
            case 171: {
                interpreter.unaryOperation(abstractInsnNode, this.pop());
                break;
            }
            case 172:
            case 173:
            case 174:
            case 175:
            case 176: {
                Value pop21 = this.pop();
                interpreter.unaryOperation(abstractInsnNode, pop21);
                interpreter.returnOperation(abstractInsnNode, pop21, this.returnValue);
                break;
            }
            case 177: {
                if (this.returnValue != null) {
                    throw new AnalyzerException(abstractInsnNode, "Incompatible return type");
                }
                break;
            }
            case 178: {
                this.push(interpreter.newOperation(abstractInsnNode));
                break;
            }
            case 179: {
                interpreter.unaryOperation(abstractInsnNode, this.pop());
                break;
            }
            case 180: {
                this.push(interpreter.unaryOperation(abstractInsnNode, this.pop()));
                break;
            }
            case 181: {
                interpreter.binaryOperation(abstractInsnNode, this.pop(), this.pop());
                break;
            }
            case 182:
            case 183:
            case 184:
            case 185: {
                ArrayList list = new ArrayList();
                String desc = ((MethodInsnNode)abstractInsnNode).desc;
                for (int i = Type.getArgumentTypes(desc).length; i > 0; --i) {
                    list.add(0, this.pop());
                }
                if (abstractInsnNode.getOpcode() != 184) {
                    list.add(0, this.pop());
                }
                if (Type.getReturnType(desc) == Type.VOID_TYPE) {
                    interpreter.naryOperation(abstractInsnNode, list);
                    break;
                }
                this.push(interpreter.naryOperation(abstractInsnNode, list));
                break;
            }
            case 186: {
                ArrayList list2 = new ArrayList();
                String desc2 = ((InvokeDynamicInsnNode)abstractInsnNode).desc;
                for (int j = Type.getArgumentTypes(desc2).length; j > 0; --j) {
                    list2.add(0, this.pop());
                }
                if (Type.getReturnType(desc2) == Type.VOID_TYPE) {
                    interpreter.naryOperation(abstractInsnNode, list2);
                    break;
                }
                this.push(interpreter.naryOperation(abstractInsnNode, list2));
                break;
            }
            case 187: {
                this.push(interpreter.newOperation(abstractInsnNode));
                break;
            }
            case 188:
            case 189:
            case 190: {
                this.push(interpreter.unaryOperation(abstractInsnNode, this.pop()));
                break;
            }
            case 191: {
                interpreter.unaryOperation(abstractInsnNode, this.pop());
                break;
            }
            case 192:
            case 193: {
                this.push(interpreter.unaryOperation(abstractInsnNode, this.pop()));
                break;
            }
            case 194:
            case 195: {
                interpreter.unaryOperation(abstractInsnNode, this.pop());
                break;
            }
            case 197: {
                ArrayList list3 = new ArrayList();
                for (int k = ((MultiANewArrayInsnNode)abstractInsnNode).dims; k > 0; --k) {
                    list3.add(0, this.pop());
                }
                this.push(interpreter.naryOperation(abstractInsnNode, list3));
                break;
            }
            case 198:
            case 199: {
                interpreter.unaryOperation(abstractInsnNode, this.pop());
                break;
            }
            default: {
                throw new RuntimeException("Illegal opcode " + abstractInsnNode.getOpcode());
            }
        }
    }
    
    public boolean merge(Frame<? extends V> frame, Interpreter<V> interpreter) throws AnalyzerException {
        if (this.top != frame.top) {
            throw new AnalyzerException(null, "Incompatible stack heights");
        }
        boolean n = false;
        for (int i = 0; i < this.locals + this.top; ++i) {
            Value merge = interpreter.merge(this.values[i], frame.values[i]);
            if (!merge.equals(this.values[i])) {
                this.values[i] = merge;
                n = true;
            }
        }
        return n;
    }
    
    public boolean merge(Frame<? extends V> frame, boolean[] array) {
        boolean n = false;
        for (int i = 0; i < this.locals; ++i) {
            if (!array[i] && !this.values[i].equals(frame.values[i])) {
                this.values[i] = frame.values[i];
                n = true;
            }
        }
        return n;
    }
    
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();
        for (int i = 0; i < this.getLocals(); ++i) {
            sb.append(this.getLocal(i));
        }
        sb.append(' ');
        for (int j = 0; j < this.getStackSize(); ++j) {
            sb.append(this.getStack(j).toString());
        }
        return sb.toString();
    }
}
